from PIL import Image
from collections import deque


def voisins_non_visites(image, noeud):
    x, y = noeud
    voisins = []
    for v_x, v_y in ((x + 1, y), (x - 1, y), (x, y + 1), (x, y - 1)):
        if 0 <= v_x < image.width and 0 <= v_y < image.height:
            if image.getpixel((v_x, v_y)) == (255, 255, 255):
                voisins.append((v_x, v_y))
    return voisins


def noeud_depart(image):
    for x in range(image.width):
        if image.getpixel((x, 0)) == (255, 255, 255):
            return (x, 0)
    raise ValueError("Pas de départ")


def dessiner_chemin(image, chemin):
    for x, y in chemin:
        image.putpixel((x, y), (255, 0, 0))


def parcours_en_profondeur(image, depart):
    noeud = depart
    chemin = []
    image.putpixel(noeud, (128, 128, 128))

    while noeud[1] < image.height - 1:
        # On cherche les voisins non visités.
        voisins = voisins_non_visites(image, noeud)
        if len(voisins) == 0:
            # On est dans un cul-de-sac.

            if len(chemin) == 0:
                # Ne devrait pas arriver si le labyrinthe est bien formé.
                return None
            
            # On revient en arrière.
            noeud = chemin.pop()
            continue
        
        # On avance.
        chemin.append(noeud)
        noeud = voisins[0]
        image.putpixel(noeud, (128, 128, 128))
    
    # On a trouvé la sortie !
    chemin.append(noeud)
    return chemin


def parcours_en_largeur(image, depart):
    noeud = depart
    file = deque()  # File des noeuds à visiter.
    file.append(noeud)  # On ajoute le noeud de départ.
    origines = {}  # Dictionnaire des origines des noeuds.

    while len(file) > 0:
        noeud = file.popleft()
        image.putpixel(noeud, (128, 128, 128))
        if noeud[1] == image.height - 1:
            # On a trouvé la sortie !

            # On reconstruit le chemin.
            chemin = []
            chemin.append(noeud)
            while noeud in origines:
                noeud = origines[noeud]
                chemin.append(noeud)
            return list(reversed(chemin))
        
        # On ajoute les voisins non visités à la file.
        for voisin in voisins_non_visites(image, noeud):
            file.append(voisin)
            # On note l'origine du voisin.
            origines[voisin] = noeud

    # Il n'y a plus de noeuds à visiter, et on n'a pas trouvé la sortie.
    return None


if __name__ == "__main__":
    # Nom du fichier à ouvrir
    nom = "labyrinthe.png"

    # Nom complet du fichier à ouvrir
    from pathlib import Path
    nom_complet = Path(__file__).with_name(nom)
    
    # Ouverture de l'image
    image = Image.open(nom_complet)

    # Recherche du chemin
    chemin = parcours_en_largeur(image, noeud_depart(image))
    dessiner_chemin(image, chemin)

    # Nom du fichier de résultat
    nom_resultat = nom[:-4] + "_resultat.png"

    # Nom complet du fichier de résultat
    nom_complet_resultat = Path(__file__).with_name(nom_resultat)

    # Sauvegarde du résultat
    image.resize((image.width * 5, image.height * 5), Image.BOX).save(nom_complet_resultat)